@@ -38,9 +38,15 @@ Formatted output (Prolog predicates format/[1,2,3]).   One  day,  the  C
 source should also use format() to produce error messages, etc.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#include "pl-incl.h"
+#include "pl-fmt.h"
 #include "pl-ctype.h"
 #include "pl-utf8.h"
+#include "../pl-arith.h"
+#include "../pl-proc.h"
+#include "../pl-fli.h"
+#include "../pl-prims.h"
+#include "../pl-write.h"
+#include "../pl-pro.h"
 #include <ctype.h>
 #include <stdio.h>
 
@@ -164,7 +170,7 @@ oututf8(format_state *state, const char *s, size_t len)
   while(s<e)
   { int chr;
 
-    s = utf8_get_char(s, &chr);
+    PL_utf8_code_point(&s, e, &chr);
     if ( !outchr(state, chr) )
       return FALSE;
   }
@@ -236,7 +242,7 @@ pl_format_predicate(term_t chr, term_t descr)
   if ( !format_predicates )
     format_predicates = newHTable(8);
 
-  addNewHTable(format_predicates, (void *)(intptr_t)c, proc);
+  updateHTable(format_predicates, (void *)(intptr_t)c, proc);
 
   succeed;
 }
@@ -295,7 +301,7 @@ format_impl(IOSTREAM *out, term_t format, term_t Args, Module m)
   int rval;
   PL_chars_t fmt;
 
-  if ( !PL_get_text(format, &fmt, CVT_ALL|BUF_RING) )
+  if ( !PL_get_text(format, &fmt, CVT_ALL|BUF_STACK) )
     return PL_error("format", 3, NULL, ERR_TYPE, ATOM_text, format);
 
   if ( (argc = (int)lengthList(args, FALSE)) >= 0 )
@@ -372,6 +378,66 @@ get_chr_from_text(const PL_chars_t *t, int index)
 }
 
 
+typedef struct sub_state
+{ char          buf[BUFSIZE];
+  char         *str;
+  size_t        bufsize;
+  format_state *fstate;
+  IOSTREAM     *old_stream;
+} sub_state;
+
+static int
+prepare_sub_format(sub_state *state, format_state *fstate, IOSTREAM *fd)
+{ state->fstate = fstate;
+
+  if ( !fstate->pending_rubber &&
+       fd->position &&
+       fd->position->linepos == fstate->column )
+  { state->old_stream = Scurout;
+
+    Scurout = fd;
+  } else
+  { state->old_stream = NULL;
+    state->str        = state->buf;
+    state->bufsize    = sizeof(state->buf);
+
+    tellString(&state->str, &state->bufsize, ENC_UTF8);
+    if ( true(fd, SIO_ISATTY) )
+      set(Scurout, SIO_ISATTY);
+  }
+
+  return TRUE;
+}
+
+static int
+end_sub_format(sub_state *state, int rc)
+{ int lp = Scurout->position->linepos;
+
+  if ( state->old_stream )
+  { Scurout = state->old_stream;
+
+    if ( rc )
+      state->fstate->column = lp;
+  } else
+  { int lp = Scurout->position->linepos;
+
+    toldString();
+    if ( rc )
+    { int c0 = state->fstate->column;
+      rc = oututf8(state->fstate, state->str, state->bufsize);
+      state->fstate->column = c0 + lp;
+    }
+    if ( state->str != state->buf )
+      free(state->str);
+  }
+
+  return rc;
+}
+
+
+
+
+
 		/********************************
 		*       ACTUAL FORMATTING	*
 		********************************/
@@ -426,7 +492,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 	    }
 	  } else if ( c == '*' )
 	  { NEED_ARG;
-	    if ( PL_get_integer(argv, &arg) )
+	    if ( PL_get_integer(argv, &arg) && arg >= 0 )
 	    { SHIFT;
 	    } else
 	      FMT_ERROR("no or negative integer for `*' argument");
@@ -446,13 +512,11 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 	       (proc = lookupHTable(format_predicates, (void*)((intptr_t)c))) )
 	  { size_t arity;
 	    term_t av;
-	    char buf[BUFSIZE];
-	    char *str = buf;
-	    size_t bufsize = BUFSIZE;
+	    sub_state sstate;
 	    int i;
 
 	    PL_predicate_info(proc, NULL, &arity, NULL);
-	    av = PL_new_term_refs(arity);
+	    av = PL_new_term_refs((int)arity);
 
 	    if ( arg == DEFAULT )
 	      PL_put_atom(av+0, ATOM_default);
@@ -465,13 +529,11 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 	      SHIFT;
 	    }
 
-	    tellString(&str, &bufsize, ENC_UTF8);
+	    if ( !(rc=prepare_sub_format(&sstate, &state, fd)) )
+	      goto out;
 	    rc = PL_call_predicate(NULL, PL_Q_PASS_EXCEPTION, proc, av);
-	    toldString();
-	    if ( rc )
-	      rc = oututf8(&state, str, bufsize);
-	    if ( str != buf )
-	      free(str);
+	    rc = end_sub_format(&sstate, rc);
+
 	    if ( !rc )
 	      goto out;
 
@@ -522,7 +584,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 		  PL_locale *l;
 
 		  NEED_ARG;
-		  if ( !valueExpression(argv, &n PASS_LD) )
+		  if ( !valueExpression(argv, &n) )
 		  { char f[2];
 
 		    f[0] = c;
@@ -556,7 +618,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 		  tmp_buffer b;
 
 		  NEED_ARG;
-		  if ( !valueExpression(argv, &i PASS_LD) ||
+		  if ( !valueExpression(argv, &i) ||
 		       !toIntegerNumber(&i, 0) )
 		  { char f[2];
 
@@ -600,7 +662,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 		  } else			/* r,R */
 		  { if ( arg == DEFAULT )
 		      FMT_ERROR("r,R requires radix specifier");
-		    if ( arg < 1 || arg > 36 )
+		    if ( arg < 2 || arg > 36 )
 		    { term_t r = PL_new_term_ref();
 
 		      PL_put_integer(r, arg);
@@ -639,8 +701,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 		  break;
 		}
 		{ Func f;
-		  char buf[BUFSIZE];
-		  char *str;
+		  sub_state sstate;
 
 	      case 'k':			/* write_canonical */
 		  f = pl_write_canonical;
@@ -656,121 +717,51 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 		  pl_common:
 
 		  NEED_ARG;
-		  if ( state.pending_rubber )
-		  { size_t bufsize = BUFSIZE;
 
-		    str = buf;
-		    tellString(&str, &bufsize, ENC_UTF8);
-		    rc = (*f)(argv);
-		    toldString();
-		    if ( rc )
-		      rc = oututf8(&state, str, bufsize);
-		    if ( str != buf )
-		      free(str);
-		    if ( !rc )
-		      goto out;
-		  } else
-		  { if ( fd->position &&
-			 fd->position->linepos == state.column )
-		    { IOSTREAM *old = Scurout;
+		  if ( !(rc=prepare_sub_format(&sstate, &state, fd)) )
+		    goto out;
+		  rc = (int)(*f)(argv);
+		  rc = end_sub_format(&sstate, rc);
 
-		      Scurout = fd;
-		      rc = (int)(*f)(argv);
-		      Scurout = old;
-		      if ( !rc )
-			goto out;
-
-		      state.column = fd->position->linepos;
-		    } else
-		    { size_t bufsize = BUFSIZE;
-
-		      str = buf;
-		      tellString(&str, &bufsize, ENC_UTF8);
-		      rc = (*f)(argv);
-		      toldString();
-		      if ( rc )
-			rc = oututf8(&state, str, bufsize);
-		      if ( str != buf )
-			free(str);
-		      if ( !rc )
-			goto out;
-		    }
-		  }
 		  SHIFT;
 		  here++;
 		  break;
 		}
 	      case 'W':			/* write_term(Value, Options) */
-	       { char buf[BUFSIZE];
-		 char *str;
+	       { sub_state sstate;
 
 		 if ( argc < 2 )
 		 { FMT_ERROR("not enough arguments");
 		 }
-		 if ( state.pending_rubber )
-		  { size_t bufsize = BUFSIZE;
 
-		    str = buf;
-		    tellString(&str, &bufsize, ENC_UTF8);
-		    rc = (int)pl_write_term(argv, argv+1);
-		    toldString();
-		    if ( rc )
-		      rc = oututf8(&state, str, bufsize);
-		    if ( str != buf )
-		      free(str);
-		    if ( !rc )
-		      goto out;
-		  } else
-		  { if ( fd->position &&
-			 fd->position->linepos == state.column )
-		    { IOSTREAM *old = Scurout;
+		 if ( !(rc=prepare_sub_format(&sstate, &state, fd)) )
+		   goto out;
+		 rc = (int)pl_write_term(argv, argv+1);
+		 rc = end_sub_format(&sstate, rc);
 
-		      Scurout = fd;
-		      rc = (int)pl_write_term(argv, argv+1);
-		      Scurout = old;
-		      if ( !rc )
-			goto out;
+		 if ( !rc )
+		   goto out;
 
-		      state.column = fd->position->linepos;
-		    } else
-		    { size_t bufsize = BUFSIZE;
-
-		      str = buf;
-		      tellString(&str, &bufsize, ENC_UTF8);
-		      rc = (int)pl_write_term(argv, argv+1);
-		      toldString();
-		      if ( rc )
-			rc = oututf8(&state, str, bufsize);
-		      if ( str != buf )
-			free(str);
-		      if ( !rc )
-			goto out;
-		    }
-		  }
-		  SHIFT;
-		  SHIFT;
-		  here++;
-		  break;
+		 SHIFT;
+		 SHIFT;
+		 here++;
+		 break;
 	       }
 	      case '@':
-	        { char buf[BUFSIZE];
-		  char *str = buf;
-		  size_t bufsize = BUFSIZE;
-		  term_t ex = 0;
+	        { term_t ex = 0;
+		  sub_state sstate;
 
 		  if ( argc < 1 )
 		  { FMT_ERROR("not enough arguments");
 		  }
-		  tellString(&str, &bufsize, ENC_UTF8);
+
+		  if ( !(rc=prepare_sub_format(&sstate, &state, fd)) )
+		    goto out;
 		  rc = callProlog(m, argv, PL_Q_CATCH_EXCEPTION, &ex);
-		  toldString();
-		  if ( rc )
-		    rc = oututf8(&state, str, bufsize);
-		  if ( str != buf )
-		    free(str);
+		  rc = end_sub_format(&sstate, rc);
 
 		  if ( !rc )
-		  { if ( ex )
+		  { if ( ex && !PL_exception(0) )
 		      rc = PL_raise_exception(ex);
 		    goto out;
 		  }
@@ -861,6 +852,8 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 
   if ( state.pending_rubber )		/* not closed ~t: flush out */
     emit_rubber(&state);
+  if ( argc != 0 )
+    FMT_ERROR("too many arguments");
 
 out:
   return rc;
@@ -912,7 +905,7 @@ emit_rubber(format_state *state)
     }
 
     if ( s < e )
-    { s = utf8_get_char(s, &chr);
+    { PL_utf8_code_point(&s, e, &chr);
       if ( Sputcode(chr, state->out) < 0 )
 	return FALSE;
     } else
@@ -1167,7 +1160,7 @@ utf8_dp(PL_locale *l, char *s, int *len)
 { if ( l->decimal_point )
   { if ( !ths_to_utf8(s, l->decimal_point, 20) )
       return FALSE;
-    *len = strlen(s);
+    *len = (int)strlen(s);
   } else
   { *s++ = '.';
     *s = EOS;
@@ -1248,7 +1241,7 @@ groupDigits(PL_locale *locale, Buffer b)
 
       if ( !ths_to_utf8(ths, locale->thousands_sep, sizeof(ths)) )
 	return FALSE;
-      thslen = strlen(ths);
+      thslen = (int)strlen(ths);
 
       if ( !growBuffer(b, thslen*groups) )
 	return PL_no_memory();
@@ -1590,6 +1583,10 @@ formatFloat(PL_locale *locale, int how, int arg, Number f, Buffer out)
 	  return NULL;
 	}
 	written = snprintf(baseBuffer(out, char), size, tmp, f->value.f);
+#ifdef __WINDOWS__
+        if ( written < 0 )	/* pre-C99 Windows snprintf() returns -1 */
+	  written = size*2;
+#endif
       }
 
 #ifdef __WINDOWS__
