@@ -3,28 +3,52 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2015, University of Amsterdam
-			      VU University Amsterdam
+    Copyright (c)  1985-2019, University of Amsterdam
+                              VU University Amsterdam
+			      CWI, Amsterdam
+    All rights reserved.
 
-    This library is free software; you can redistribute it and/or
-    modify it under the terms of the GNU Lesser General Public
-    License as published by the Free Software Foundation; either
-    version 2.1 of the License, or (at your option) any later version.
+    Redistribution and use in source and binary forms, with or without
+    modification, are permitted provided that the following conditions
+    are met:
 
-    This library is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-    Lesser General Public License for more details.
+    1. Redistributions of source code must retain the above copyright
+       notice, this list of conditions and the following disclaimer.
 
-    You should have received a copy of the GNU Lesser General Public
-    License along with this library; if not, write to the Free Software
-    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
+    2. Redistributions in binary form must reproduce the above copyright
+       notice, this list of conditions and the following disclaimer in
+       the documentation and/or other materials provided with the
+       distribution.
+
+    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
+    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
+    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
+    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
+    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
+    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
+    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
+    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+    POSSIBILITY OF SUCH DAMAGE.
 */
 
 #ifdef SECURE_GC
 #define O_DEBUG 1			/* include checkData() */
 #endif
-#include "pl-incl.h"
+#include "pl-pro.h"
+#include "pl-arith.h"
+#include "os/pl-cstack.h"
+#include "os/pl-ctype.h"
+#include "pl-event.h"
+#include "pl-tabling.h"
+#include "pl-trace.h"
+#include "pl-setup.h"
+#include "pl-wam.h"
+#include "pl-attvar.h"
+#include "pl-proc.h"
+#include "pl-fli.h"
 
 
 		/********************************
@@ -58,15 +82,35 @@ resetProlog(int clear_stacks)
   }
 
 #ifdef O_LIMIT_DEPTH
-  depth_limit   = (uintptr_t)DEPTH_NO_LIMIT;
+  LD->depth_info.limit = DEPTH_NO_LIMIT;
 #endif
 
+  LD->autoload.nesting = NULL;
+  LD->autoload.loop = NULL;
   updateAlerted(LD);
 
   return TRUE;
 }
 
 
+static int
+restore_after_exception(term_t except)
+{ GET_LD
+  atom_t a;
+  int rc = TRUE;
+
+  tracemode(FALSE, NULL);
+  debugmode(DBG_OFF, NULL);
+  setPrologFlagMask(PLFLAG_LASTCALL);
+  if ( PL_get_atom(except, &a) && a == ATOM_aborted )
+  { rc = ( callEventHook(PLEV_ABORT) &&
+	   printMessage(ATOM_informational, PL_ATOM, ATOM_aborted) );
+  }
+
+  return rc;
+}
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 query_loop() runs a zero-argument goal on   behalf  of the toplevel. The
 reason for this to be in C is to   be able to handle exceptions that are
@@ -74,7 +118,7 @@ considered unhandled and thus  can  trap   the  debugger.  I.e., if goal
 terminates due to an exception, the exception   is  reported and goal is
 restarted. Before the restart, the system is   restored to a sane state.
 This notably affects I/O  (reset  current  I/O   to  user  I/O)  and the
-debugger.
+debugger.  Return: FALSE: failed, TRUE: success, -1: exception.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 int
@@ -83,6 +127,9 @@ query_loop(atom_t goal, int loop)
   int rc;
   int clear_stacks = (LD->query == NULL);
 
+  if ( loop )
+    enable_debug_on_interrupt(TRUE);
+
   do
   { fid_t fid;
     qid_t qid = 0;
@@ -101,28 +148,32 @@ query_loop(atom_t goal, int loop)
     } else
     { error:
       except = exception_term;
-      rc = FALSE;			/* Won't get any better */
+      rc = -1;				/* Won't get any better */
       break;
     }
 
     if ( !rc && (except = PL_exception(qid)) )
-    { atom_t a;
-
-      tracemode(FALSE, NULL);
-      debugmode(DBG_OFF, NULL);
-      setPrologFlagMask(PLFLAG_LASTCALL);
-      if ( PL_get_atom(except, &a) && a == ATOM_aborted )
-      { callEventHook(PLEV_ABORT);
-        printMessage(ATOM_informational, PL_ATOM, ATOM_aborted);
-      }
+    { restore_after_exception(except);
+      rc = -1;
     }
 
     if ( qid ) PL_close_query(qid);
     if ( fid ) PL_discard_foreign_frame(fid);
     if ( !except )
       break;
+#ifdef O_PLMT
+    if (LD->exit_requested)
+      loop = 0;
+#endif
   } while(loop);
 
+#ifdef O_PLMT
+  DEBUG(MSG_CLEANUP_THREAD,
+	if ( LD->exit_requested )
+	Sdprintf("Thread %d: leaving REPL loop due to exit_requested\n",
+		 PL_thread_self()));
+#endif
+
   return rc;
 }
 
@@ -136,40 +187,43 @@ the debugger.  Restores I/O and debugger on exit.  The Prolog  predicate
 static int
 pl_break1(atom_t goal)
 { GET_LD
-  int rc;
+  int rc = TRUE;
   int old_level = LD->break_level;
 
-  IOSTREAM *inSave  = Scurin;
-  IOSTREAM *outSave = Scurout;
-  intptr_t skipSave = debugstatus.skiplevel;
-  int  suspSave     = debugstatus.suspendTrace;
+  IOSTREAM *inSave       = Scurin;
+  IOSTREAM *outSave      = Scurout;
+  intptr_t skipSave      = debugstatus.skiplevel;
+  int  suspSave          = debugstatus.suspendTrace;
   int  traceSave;
   debug_type debugSave;
+  tbl_status tblstat;
 
   tracemode(FALSE, &traceSave);
   debugmode(DBG_OFF, &debugSave);
+  save_tabling_status(&tblstat);
 
   Scurin  = Sinput;
   Scurout = Soutput;
 
   LD->break_level++;
   if ( LD->break_level > 0 )
-  { printMessage(ATOM_informational,
-		 PL_FUNCTOR, FUNCTOR_break2,
-	           PL_ATOM, ATOM_begin,
-		   PL_INT,  LD->break_level);
+  { rc = printMessage(ATOM_informational,
+		      PL_FUNCTOR, FUNCTOR_break2,
+		        PL_ATOM, ATOM_begin,
+		        PL_INT,  LD->break_level);
   }
 
-  rc = query_loop(goal, TRUE);
+  rc = rc && (query_loop(goal, TRUE) == TRUE);
 
   if ( LD->break_level > 0 )
-  { printMessage(ATOM_informational,
-		 PL_FUNCTOR, FUNCTOR_break2,
-	           PL_ATOM, ATOM_end,
-		   PL_INT,  LD->break_level);
+  { rc = rc && printMessage(ATOM_informational,
+			    PL_FUNCTOR, FUNCTOR_break2,
+			      PL_ATOM, ATOM_end,
+			      PL_INT,  LD->break_level);
   }
   LD->break_level = old_level;
 
+  restore_tabling_status(&tblstat);
   debugstatus.suspendTrace = suspSave;
   debugstatus.skiplevel    = skipSave;
   tracemode(traceSave, NULL);
@@ -193,11 +247,11 @@ pl_break(void)
 { GET_LD
   wakeup_state wstate;
 
-  if ( saveWakeup(&wstate, TRUE PASS_LD) )
+  if ( saveWakeup(&wstate, TRUE) )
   { word rc;
 
     rc = pl_break1(ATOM_dquery_loop);
-    restoreWakeup(&wstate PASS_LD);
+    restoreWakeup(&wstate);
 
     return rc;
   }
@@ -248,21 +302,40 @@ Execute Goal as once/1 while blocking signals.
 static
 PRED_IMPL("$sig_atomic", 1, sig_atomic, PL_FA_TRANSPARENT)
 { PRED_LD
-  term_t ex;
   int rval;
 
   startCritical;
-  rval = callProlog(NULL, A1, PL_Q_CATCH_EXCEPTION, &ex);
+  rval = callProlog(NULL, A1, PL_Q_PASS_EXCEPTION, NULL);
   if ( !endCritical )
     fail;				/* aborted */
 
-  if ( !rval && ex )
-    return PL_raise_exception(ex);
-
   return rval;
 }
 
 
+/** '$call_no_catch'(:Goal)
+ *
+ * Runs a goal for the toplevel.  This notably means that exceptions
+ * are considered _uncaught_, are printed and ignored.  Also the
+ * truth value is ignored.
+ */
+
+static
+PRED_IMPL("$call_no_catch", 1, call_no_catch, PL_FA_TRANSPARENT)
+{ int rc;
+  term_t ex;
+
+  rc = callProlog(NULL, A1, PL_Q_NORMAL, &ex);
+  if ( !rc && ex )
+  { restore_after_exception(ex);
+    PL_clear_exception();
+  }
+
+  return TRUE;
+}
+
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Call a prolog goal from C. The argument must  be  an  instantiated  term
 like for the Prolog predicate call/1.
@@ -304,6 +377,12 @@ callProlog(Module module, term_t goal, int flags, term_t *ex)
     fail;
   }
 
+  /* Quick check for '$sig_atomic'(true) resulting from the call_cleanup
+   * series
+   */
+  if ( fd == FUNCTOR_true0 )
+    return TRUE;
+
   proc = resolveProcedure(fd, module);
 
   { int arity = arityFunctor(fd);
@@ -335,6 +414,16 @@ callProlog(Module module, term_t goal, int flags, term_t *ex)
     }
 
     PL_cut_query(qid);
+    if ( rval && exception_term )
+    { if ( ex )
+      { PL_put_term(ex2, exception_term);
+	PL_clear_exception();
+	*ex = ex2;
+	reset = g;
+      }
+      rval = FALSE;
+    }
+
     if ( !reset )
       reset = (ex2 ? ex2 : g);
     PL_reset_term_refs(reset);
@@ -366,6 +455,10 @@ abortProlog(void)
     PL_close_foreign_frame(fid);
   }
 
+#ifdef O_MAINTENANCE
+  save_backtrace("abort");
+#endif
+
   return rc;
 }
 
@@ -403,7 +496,7 @@ prologToplevel(atom_t goal)
   rc = query_loop(goal, loop);
   LD->break_level = old_level;
 
-  return rc;
+  return rc == TRUE;
 }
 
 
@@ -452,11 +545,17 @@ those during normal execution.  Arity of terms is limited to  100  as  a
 kind of heuristic.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
+typedef struct
+{ int recursive;
+  int errors;
+  int flags;
+} chk_data;
+
 #define onGlobal(p) onStack(global, p)
 #define onLocal(p) onStack(local, p)
 
 static void
-printk(char *fm, ...)
+printk(chk_data *context, char *fm, ...)
 { va_list args;
 
   va_start(args, fm);
@@ -465,7 +564,8 @@ printk(char *fm, ...)
   Sfprintf(Serror, "]\n");
   va_end(args);
 
-  assert(0);
+  context->errors++;
+  trap_gdb();
 }
 
 static intptr_t check_marked;
@@ -473,8 +573,9 @@ static intptr_t check_marked;
 #define mark(p)		(*(p) |= MARK_MASK, check_marked++)
 #define unmark(p)	(*(p) &= ~MARK_MASK, check_marked--)
 
+#define unmark_data(p) LDFUNC(unmark_data, p)
 static void
-unmark_data(Word p ARG_LD)
+unmark_data(DECL_LD Word p)
 {
 last_arg:
   deRef(p);
@@ -491,7 +592,7 @@ last_arg:
     f = valueTerm(*p);
     arity = arityFunctor(f->definition);
     for(n=0; n<arity-1; n++)
-      unmark_data(&f->arguments[n] PASS_LD);
+      unmark_data(&f->arguments[n]);
     p = &f->arguments[n];
     goto last_arg;
   }
@@ -510,8 +611,9 @@ is_ht_capacity(int arity)
 }
 #endif
 
+#define check_data(p, context) LDFUNC(check_data, p, context)
 static word
-check_data(Word p, int *recursive ARG_LD)
+check_data(DECL_LD Word p, chk_data *context)
 { int arity; int n;
   Word p2;
   word key = 0L;
@@ -528,13 +630,13 @@ last_arg:
               if ( !isAttVar(*p2) )
 #endif
                 if ( !gc_status.blocked )
-                  printk("Reference to higher address");
+                  printk(context, "Reference to higher address");
             }
           });
     if ( p2 == p )
-      printk("Reference to same address");
+      printk(context, "Reference to same address");
     if ( !onLocal(p2) && !onGlobal(p2) )
-      printk("Illegal reference pointer at 0x%x --> 0x%x", p, p2);
+      printk(context, "Illegal reference pointer at %p --> %p", p, p2);
 
     p = p2;
   }
@@ -545,7 +647,7 @@ last_arg:
 #ifdef O_ATTVAR
   if ( isAttVar(*p) )
   { if ( is_marked(p) )			/* loop */
-    { (*recursive)++;
+    { context->recursive++;
       return key;
     }
 
@@ -555,11 +657,13 @@ last_arg:
 
 					/* See argument_stack_to_term_refs() */
     if ( !onGlobal(p) && (!gc_status.active || p < (Word)environment_frame) )
-      printk("attvar: not on global stack: 0x%x", p);
+      printk(context, "attvar: not on global stack: %p", p);
     if ( !onGlobal(p2) )
-      printk("attvar: attribute not on global stack: 0x%x --> 0x%x", p, p2);
+      printk(context, "attvar: attribute not on global stack: %p --> %p", p, p2);
     if ( p == p2 )
-      printk("attvar: self-reference: 0x%x", p);
+      printk(context, "attvar: self-reference: %p", p);
+    if ( !(context->flags&CHK_DATA_NOATTVAR_CHAIN) && !on_attvar_chain(p) )
+      printk(context, "attvar: not on attvar chain: %p", p);
 
     p = p2;
     goto last_arg;
@@ -577,9 +681,9 @@ last_arg:
     assert(!is_marked(p));
 
     if ( !onGlobal(a) )
-      printk("Indirect at %p not on global stack", a);
+      printk(context, "Indirect at %p not on global stack", a);
     if ( storage(*p) != STG_GLOBAL )
-      printk("Indirect data not on global");
+      printk(context, "Indirect data not on global");
     if ( isBignum(*p) )
       return key+(word) valBignum(*p);
     if ( isFloat(*p) )
@@ -593,11 +697,11 @@ last_arg:
 
 	if ( sz != (len=strlen(s)) )
 	{ if ( sz < len )
-	    printk("String has inconsistent length: 0x%x", *p);
+	    printk(context, "String has inconsistent length: %p", *p);
 	  else if ( s[sz] )
-	    printk("String not followed by NUL-char: 0x%x", *p);
+	    printk(context, "String not followed by NUL-char: %p", *p);
 /*	else
-	    printf("String contains NUL-chars: 0x%x", *p);
+	    printf("String contains NUL-chars: %p", *p);
 */
 	}
       } else
@@ -608,9 +712,9 @@ last_arg:
 
 	if ( sz != (len=wcslen(s)) )
 	{ if ( sz < len )
-	    printk("String has inconsistent length: 0x%x", *p);
+	    printk(context, "String has inconsistent length: %p", *p);
 	  else if ( s[sz] )
-	    printk("String not followed by NUL-char: 0x%x", *p);
+	    printk(context, "String not followed by NUL-char: %p", *p);
 	}
       }
       return key + *addressIndirect(*p);
@@ -618,8 +722,10 @@ last_arg:
 #ifdef O_GMP
     if ( isMPZNum(*p) )
       return 0x62f8da3c;		/* TBD: make key from MPZ */
+    if ( isMPQNum(*p) )
+      return 0xed7ef4ea;		/* TBD: make key from MPQ */
 #endif
-    printk("Illegal indirect datatype");
+    printk(context, "Illegal indirect datatype");
     return key;
   }
 
@@ -629,11 +735,11 @@ last_arg:
 
     assert(!is_marked(p));
     if ( storage(*p) != STG_STATIC )
-      printk("Atom doesn't have STG_STATIC");
+      printk(context, "Atom doesn't have STG_STATIC");
 
     idx = indexAtom(*p);
     if ( idx >= mx )
-      printk("Atom index out of range (%ld > %ld)", idx, mx);
+      printk(context, "Atom index out of range (%ld > %ld)", idx, mx);
     return key + *p;
   }
   if ( tagex(*p) == (TAG_VAR|STG_RESERVED) )
@@ -642,39 +748,39 @@ last_arg:
 					/* now it should be a term */
   if ( tag(*p) != TAG_COMPOUND ||
        storage(*p) != STG_GLOBAL )
-    printk("Illegal term at: %p: 0x%x", p, *p);
+    printk(context, "Illegal term at: %p: %p", p, *p);
 
   if ( is_marked(p) )
-  { (*recursive)++;
+  { context->recursive++;
     return key;				/* recursive */
   }
 
   { Functor f = valueTerm(*p);
 
     if ( !onGlobal(f) )
-      printk("Term at %p not on global stack", f);
+      printk(context, "Term at %p not on global stack", f);
 
     if ( tag(f->definition) != TAG_ATOM ||
          storage(f->definition) != STG_GLOBAL )
-      printk("Illegal functor: 0x%x", *p);
+      printk(context, "Illegal functor: %p", *p);
     if ( f->definition & MARK_MASK )
-      printk("functor with mark: 0x%x", *p);
+      printk(context, "functor with mark: %p", *p);
     if ( f->definition & FIRST_MASK )
-      printk("functor with first: 0x%x", *p);
+      printk(context, "functor with first: %p", *p);
     arity = arityFunctor(f->definition);
     if ( arity < 0 )
-      printk("Illegal arity (%d)", arity);
+      printk(context, "Illegal arity (%d)", arity);
     else if ( arity == 0 )
       return key;
     else
       DEBUG(CHK_HIGH_ARITY,
             { if ( arity > 256 && !is_ht_capacity(arity) )
-                printk("Dubious arity (%d)", arity);
+                printk(context, "Dubious arity (%d)", arity);
             });
 
     mark(p);
     for(n=0; n<arity-1; n++)
-      key += check_data(&f->arguments[n], recursive PASS_LD);
+      key += check_data(&f->arguments[n], context);
 
     p = &f->arguments[n];
     goto last_arg;
@@ -683,17 +789,23 @@ last_arg:
 
 
 word
-checkData(Word p)
+checkDataEx(Word p, int flags)
 { GET_LD
-  int recursive = 0;
+  chk_data context = {0};
   word key;
 
-  key = check_data(p, &recursive PASS_LD);
-  unmark_data(p PASS_LD);
+  context.flags = flags;
+  key = check_data(p, &context);
+  unmark_data(p);
 
   return key;
 }
 
+word
+checkData(Word p)
+{ return checkDataEx(p, 0);
+}
+
 #endif /* TEST */
 
 int
@@ -718,7 +830,7 @@ PL_check_data(term_t data)
    MacOS Lion.  Called from pl-vmi.c in I_EXITCLEANUP.
 */
 
-#ifdef __llvm__
+#if defined(__llvm__) && defined(VMCODE_IS_ADDRESS)
 int
 llvm_dummy(void)
 { return 0;
@@ -767,4 +879,5 @@ BeginPredDefs(pro)
   PRED_DEF("notrace", 1, notrace, PL_FA_TRANSPARENT|PL_FA_NOTRACE)
   PRED_DEF("$sig_atomic", 1, sig_atomic, PL_FA_TRANSPARENT)
   PRED_DEF("$trap_gdb", 0, trap_gdb, 0)
+  PRED_DEF("$call_no_catch", 1, call_no_catch, PL_FA_TRANSPARENT)
 EndPredDefs
